/* one-directional tcp pipe tool.

Copyright (c) 2003 - 2005 Heikki Orsila <heikki.orsila@iki.fi>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>

#include "net.h"
#include "pipe.h"
#include "version.h"
#include "accounting.h"


static int listen_socket(char *bind_host, char *port)
{
  int listen_fd;
  char printed_id[256];
  struct net_host_id id;
  int socklen;
  char sockbuf[256];
  struct sockaddr_in *sockname = (struct sockaddr_in *) sockbuf;
  int ret = -1;
  
  listen_fd = net_bind(NULL, bind_host, port, "tcp", 1 /* REUSE */);
  if (listen_fd < 0) {
    fprintf(stderr, "can't bind server port\n");
    exit(-1);
  }
  if (listen(listen_fd, 5)) {
    perror("can not listen server port");
    exit(-1);
  }
  
  while (1) {
    socklen = sizeof(sockbuf);
    ret = accept(listen_fd, (struct sockaddr *) sockname, &socklen);
    if (ret < 0) {
      if (errno == EAGAIN || errno == EINTR)
	continue;
      perror("client accept failed");
      exit(-1);
    }
    break;
  }
  
  close(listen_fd);

  if (!net_get_id(&id, (struct sockaddr *) sockname)) {
    fprintf(stderr, "client connected. can't get info on socket. rejecting.\n");
    exit(-1);
  }
  
  printed_id[0] = 0;
  net_print_id(printed_id, &id, sizeof(printed_id));
  fprintf(stderr, "client connected from %s\n", printed_id);
  return ret;
}


static int open_socket(char *net_host, char *net_port, char *bind_host)
{
  int ret = net_open(net_host, net_port, "tcp", bind_host);
  if (ret < 0) {
    fprintf(stderr, "couldn't connect to %s:%s\n", net_host, net_port);
    exit(-1);
  }
  return ret;
}


static void print_help(int only_version)
{
  fprintf(stderr, "tcppipe %s by Heikki Orsila <heikki.orsila@iki.fi>\n", TOOL_VERSION);
  fprintf(stderr, "\n");
  if (only_version)
    return;
  fprintf(stderr, "DESCRIPTION: tcppipe is a TCP pipe tool for one-directional traffic. It can do\n");
  fprintf(stderr, "             simple network load balancing by accounting transfers.\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "USAGE: tcppipe [-b file] [-i secs] [-l rate] [-n file] [-p] SOURCE DESTINATION\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "Both SOURCE and ADDRESS can be selected from following alternatives:\n");
  fprintf(stderr, "\n");
  fprintf(stderr, " '-'         means stdin for SOURCE and stdout for DESTINATION\n");
  fprintf(stderr, " 'host port' means network host and its port. The host can be just '.' and\n");
  fprintf(stderr, "             in that case host is selected from a host file given with -n.\n");
  fprintf(stderr, " ':port'     means listening socket on a port\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "Both SOURCE and DESTINATION may not be sockets.\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "OPTIONS:\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "-b/--bind-file file    Select local socket address (interface) from a bind\n");
  fprintf(stderr, "                       host file. The interface with least amount of\n");
  fprintf(stderr, "                       transfered bytes is selected.\n");
  fprintf(stderr, "-h                     Print help\n");
  fprintf(stderr, "-i secs                Set host file update interval in seconds. Default 512s.\n");
  fprintf(stderr, "-l/--limit rate        Limit network throughput to a maximum number of bytes\n");
  fprintf(stderr, "                       per second.\n");
  fprintf(stderr, "-n/--network-file file Select network host from a network host file. The host\n");
  fprintf(stderr, "                       with least amount of transfered bytes is selected.\n");
  fprintf(stderr, "-p                     Print network throughput\n");
  fprintf(stderr, "-r/--rate rate         Same as -l / --limit\n");
  fprintf(stderr, "-v/--version           Print version information.\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "EXAMPLES:\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "1. Bind to port 2345 on an interface with least amount of transfers. And direct\n");
  fprintf(stderr, "   incoming TCP connection to stdout.\n");
  fprintf(stderr, "       tcppipe -b file :2345 -\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "2. Tar some files to a pipe, and forward that data to a host using the local\n");
  fprintf(stderr, "   interface with least amount of transfers.\n");
  fprintf(stderr, "       tar cvf - *FILES* |tcppipe -b file - host port\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "3. Send some files to the host with least amount of transfers and limit rate\n");
  fprintf(stderr, "   to 100 KiB/s.\n");
  fprintf(stderr, "       tar cvf - *FILES* |tcppipe -n hostfile -l 102400 - . port\n");
  fprintf(stderr, "\n");
}


int main(int argc, char *argv[])
{
  int i;
  int src_fd, dst_fd;
  char *bind_host = NULL;
  char *net_host = NULL;
  char *net_port = NULL;
  char *bind_accounting_file = NULL;
  char *net_accounting_file = NULL;

  char *first_host = NULL;
  char *first_file = NULL;
  char *second_host = NULL;
  char *second_file = NULL;

  char buf[1024];

  struct pipe p1;

  enum {
    A_STDIO = 1,
    A_SOCKET,
    A_LSOCKET
  };

  int src_type; /* 0 = stdin, 1 = socket, 2 = listening socket */
  int dst_type; /* 0 = stdout, 1 = socket, 2 = listening socket */

  pmr_shut_up = 1;
  pmr_print_pid = 0;

  for (i = 1; i < argc;) {

    if (!strcmp("-h", argv[i]) || !strcmp("--help", argv[i])) {
      print_help(0);
      return 0;

    } else if (!strcmp("-b", argv[i]) || !strcmp("--bind-file", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give host file for -b\n");
	print_help(0);
	return -1;
      }
      bind_accounting_file = strdup(argv[i + 1]);
      if (bind_accounting_file == NULL) {
	fprintf(stderr, "not enough memory for host filename\n");
	return -1;
      }
      i += 2;

    } else if (!strcmp("-i", argv[i]) || !strcmp("--interval", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "you must provide the interval value\n");
	return -1;
      }
      pmr_accounting_interval = atoi(argv[i + 1]);
      if (pmr_accounting_interval <= 0) {
	fprintf(stderr, "illegal accounting interval: %s\n", argv[i + 1]);
	return -1;
      }
      i += 2;

    } else if (!strcmp("-l", argv[i]) || !strcmp("--limit", argv[i]) ||
	       !strcmp("-r", argv[i]) || !strcmp("--rate", argv[i]))    {
      if ((i + 1) < argc) {
        pmr_max_rate = atoi(argv[i + 1]);
	if (pmr_max_rate <= 0) {
	  fprintf (stderr, "illegal bytes per second value (%d)\n", pmr_max_rate);
	  return -1;
	}
      } else {
        fprintf (stderr, "expecting a rate limit value (bytes per second) with -r\n");
	print_help(0);
        return -1;
      }
      i += 2;

    } else if (!strcmp("-n", argv[i]) || !strcmp("--network-file", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give host file for -n\n");
	print_help(0);
	return -1;
      }
      net_accounting_file = strdup(argv[i + 1]);
      if (net_accounting_file == NULL) {
	fprintf(stderr, "not enough memory for host filename\n");
	return -1;
      }
      i += 2;

    } else if (!strcmp("-p", argv[i]) || !strcmp("--print-rate\n", argv[i])) {
      pmr_shut_up = 0;
      i++;

    } else if (!strcmp("-v", argv[i]) || !strcmp("--version", argv[i])) {
      print_help(1);
      return 0;

    } else if (!strcmp("-", argv[i])) {
      break;

    } else if (argv[i][0] == '-') {
      fprintf(stderr, "illegal args (%s)\n", argv[i]);
      print_help(0);
      return -1;
    } else {
      break;
    }
  }

  if (i >= argc) {
    if (argc != 1)
      fprintf(stderr, "You must give some arguments..\n\n");
    print_help(0);
    return 0;
  }
  if (strcmp(argv[i], "-") == 0) {
    src_type = A_STDIO;
    i++;
    goto getdest;
  }
  if (argv[i][0] == ':') {
    src_type = A_LSOCKET;
    net_port = argv[i] + 1;
    i++;
    goto getdest;
  }
  if ((i + 1) >= argc) {
    fprintf(stderr, "source network address missing\n");
    return -1;
  }
  src_type = A_SOCKET;
  net_host = argv[i];
  net_port = argv[i + 1];
  i += 2;

 getdest:
  if (i >= argc) {
    fprintf(stderr, "destination address missing\n");
    return -1;
  }
  if (strcmp(argv[i], "-") == 0) {
    dst_type = A_STDIO;
    i++;
    goto action;
  }
  if (argv[i][0] == ':') {
    dst_type = A_LSOCKET;
    net_port = argv[i] + 1;
    i++;
    goto action;
  }
  if ((i + 1) >= argc) {
    fprintf(stderr, "destination network address missing\n");
    return -1;
  }
  dst_type = A_SOCKET;
  net_host = argv[i];
  net_port = argv[i + 1];
  i += 2;

 action:

  if (i < argc) {
    fprintf(stderr, "unexpected extra argument given: %s\n", argv[i]);
    return -1;
  }

  if ((src_type == A_SOCKET || src_type == A_LSOCKET) &&
      (dst_type == A_SOCKET || dst_type == A_LSOCKET)) {
    fprintf(stderr, "only the other address can be a socket, not both.\n");
    return -1;
  }

  if (net_accounting_file != NULL) {
    long long bytes;
    if ((bytes = get_account(buf, sizeof(buf), net_accounting_file)) < 0) {
      fprintf(stderr, "no hosts found from network host file %s\n", net_accounting_file);
      return -1;
    }

    fprintf(stderr, "choosing %s as the network host with %.1f mebibytes (%lld bytes)\n", buf, ((double) bytes) / (1024.0 * 1024.0), bytes);

    net_host = strdup(buf);
    if (net_host == NULL) {
      fprintf(stderr, "no memory for net_host\n");
      return -1;
    }
  }

  if (bind_accounting_file != NULL) {
    long long bytes;
    if ((bytes = get_account(buf, sizeof(buf), bind_accounting_file)) < 0) {
      fprintf(stderr, "no hosts found from bind host file %s\n", bind_accounting_file);
      return -1;
    }
    fprintf(stderr, "choosing %s as the bind interface with %.1f mebibytes (%lld bytes)\n", buf, ((double) bytes) / (1024.0 * 1024.0), bytes);
    bind_host = strdup(buf);
    if (bind_host == NULL) {
      fprintf(stderr, "no memory for bind host\n");
      return -1;
    }
  }

  src_fd = -1;
  switch (src_type) {
  case A_STDIO:
    src_fd = 0;
    break;
  case A_LSOCKET:
    src_fd = listen_socket(bind_host, net_port);
    break;
  case A_SOCKET:
    src_fd = open_socket(net_host, net_port, bind_host);
    break;
  }

  dst_fd = -1;
  switch (dst_type) {
  case A_STDIO:
    dst_fd = 1;
    break;
  case A_LSOCKET:
    dst_fd = listen_socket(bind_host, net_port);
    break;
  case A_SOCKET:
    dst_fd = open_socket(net_host, net_port, bind_host);
    break;
  }

  p1 = (struct pipe) {.src_fd = src_fd, .dst_fd = dst_fd};

  if (net_accounting_file != NULL && bind_accounting_file != NULL) {
    first_host = net_host;
    first_file = net_accounting_file;
    second_host = bind_host;
    second_file = bind_accounting_file;
  } else if (net_accounting_file != NULL) {
    first_host = net_host;
    first_file = net_accounting_file;
  } else if (bind_accounting_file != NULL) {
    first_host = bind_host;
    first_file = bind_accounting_file;
  }

  /* here is the main loop copying data from one socket to another */
  pmr_loop(&p1, NULL, first_host, first_file, second_host, second_file);

  close(p1.src_fd);
  close(p1.dst_fd);

  if (bind_accounting_file != NULL) {
    long long bytes;
    long long incr = p1.accbytes;
    if ((bytes = account_increment(bind_host, bind_accounting_file, incr)) < 0) {
      fprintf(stderr, "%s not found from host file. can not increment!\n", bind_host);
      exit(-1);
    }
    fprintf(stderr, "host %s incremented by %.1f mebibytes to %.1f mebibytes (%lld bytes)\n", bind_host, ((double) incr) / (1024.0 * 1024.0), ((double) bytes) / (1024.0 * 1024.0), bytes);
  }

  if (net_accounting_file != NULL) {
    long long bytes;
    long long incr = p1.accbytes;
    if ((bytes = account_increment(net_host, net_accounting_file, incr)) < 0) {
      fprintf(stderr, "%s not found from host file. can not increment!\n", net_host);
      exit(-1);
    }
    fprintf(stderr, "host %s incremented by %.1f mebibytes to %.1f mebibytes (%lld bytes)\n", net_host, ((double) incr) / (1024.0 * 1024.0), ((double) bytes) / (1024.0 * 1024.0), bytes);
  }

  return 0;
}
